@@ -1,6 +1,6 @@
 ;;; examples/nyacc/lang/c99/ffi-help.scm
 
-;; Copyright (C) 2016-2020 Matthew R. Wette
+;; Copyright (C) 2016-2021 Matthew R. Wette
 ;;
 ;; This library is free software; you can redistribute it and/or
 ;; modify it under the terms of the GNU Lesser General Public
@@ -81,7 +81,7 @@
   #:use-module (sxml match)
   #:use-module ((sxml xpath)
 		#:renamer (lambda (s) (if (eq? s 'filter) 'sxml:filter s)))
-  #:use-module (srfi srfi-1)
+  #:use-module ((srfi srfi-1) #:select (fold fold-right remove last))
   #:use-module (srfi srfi-11)
   #:use-module (srfi srfi-37)
   #:use-module (system base pmatch)
@@ -92,7 +92,7 @@
   #:use-module (ice-9 regex)
   #:use-module (ice-9 pretty-print)
   #:re-export (*nyacc-version*)
-  #:version (1 02 1))
+  #:version (1 04 1))
 
 (define fh-cpp-defs
   (cond
@@ -128,6 +128,12 @@
 
 (define *errmsgs* (make-parameter '()))	; list of warnings
 
+(define dev/null
+  (let ((port #f))
+    (lambda ()
+      (unless port (set! port (open-output-file "/dev/null")))
+      port)))
+
 (define (sfscm fmt . args)
   (apply simple-format (*mport*) fmt args))
 (define* (ppscm tree #:key (per-line-prefix ""))
@@ -275,9 +281,9 @@
 (define (sw/*-desc name) (string-append name "*-desc"))
 (define (sw/& name) (string-append name "&"))
 (define (sw/struct name) (string-append "struct-" name))
-(define (sw/union name) (string-append "struct-" name))
+(define (sw/union name) (string-append "union-" name))
 (define (sw/struct* name) (string-append "struct-" name "*"))
-(define (sw/union* name) (string-append "struct-" name "*"))
+(define (sw/union* name) (string-append "union-" name "*"))
 
 ;; I was using (pointer . name) in *defined* but this has issues
 ;; because expand-typerefs does not recognize it.  Is a change needed?
@@ -301,7 +307,8 @@
 		     (if pkg-config (pkg-config-libs pkg-config) '())
 		     libraries))
  	 (libraries (delete "libm" libraries))) ;; workaround for ubuntu
-    (sfscm ";; generated with `guild compile-ffi ~A.ffi'\n" (m-path->f-path path))
+    (sfscm ";; generated with `guild compile-ffi ~A.ffi'\n"
+	   (m-path->f-path path))
     (nlscm)
     (sfscm "(define-module ~S\n" path)
     (for-each
@@ -322,7 +329,8 @@
     ;;
     (ppscm
      `(define ,(link-libs)
-	(list ,@(map (lambda (l) `(dynamic-link ,l)) (reverse libraries)))))
+	(delay
+	  (list ,@(map (lambda (l) `(dynamic-link ,l)) (reverse libraries))))))
     (sfscm "\n")
     (if (*echo-decls*) (sfscm "(define echo-decls #t)\n\n"))
     ;;
@@ -386,7 +394,8 @@
 
 ;; just the type, so parent has to build the name-value pairs for
 ;; struct members
-(define (mtail->bs-desc mdecl-tail)
+;;(define (mtail->bs-desc mdecl-tail)
+(define-public (mtail->bs-desc mdecl-tail)
   (let ((defined (*defined*))) ;; (udict (*udict*)))
     (pmatch mdecl-tail
       ;; expand typeref, use renamer,
@@ -468,7 +477,7 @@
       (((array-of ,n) . ,rest)
        `(bs:vector ,(const-expr->number n) ,(mtail->bs-desc rest)))
       (((array-of) . ,rest)
-       `(fh:pointer ,(mtail->bs-desc rest)))
+       `(bs:vector 0 ,(mtail->bs-desc rest)))
 
       (((bit-field ,size) . ,rest)
        `(bit-field ,(const-expr->number size) ,(mtail->bs-desc rest)))
@@ -571,9 +580,6 @@
 
 ;; --- structs and unions
 
-;; scheme-bytestructures will be adding
-;;   bounding-struct-descriptor union-desc => struct-desc
-
 ;; This routine will munge the fields and then perform typeref expansion.
 ;; `defined' here means has -desc (what?)
 (define (expand-field-list-typerefs field-list)
@@ -581,11 +587,12 @@
     (cons 'field-list
 	  (fold-right
 	   (lambda (pair seed)
-	     (cons (expand-typerefs (cdr pair) udict defined) seed))
-	   '() (fold-right unitize-comp-decl '() (cdr field-list))))))
+	     (cons (expand-typerefs pair udict defined) seed))
+	   '() (clean-and-unitize-fields (cdr field-list))))))
 
 ;; field-list is (field-list . ,fields)
-(define (cnvt-field-list field-list)
+;;(define (cnvt-field-list field-list)
+(define-public (cnvt-field-list field-list)
 
   (define (acons-defn name type seed)
     (cons (eval-string (simple-format #f "(quote `(~A ,~S))" name type)) seed))
@@ -595,8 +602,17 @@
       (cons (eval-string
 	     (simple-format #f "(quote `(~A ,~S ~A))" name type size)) seed)))
 
-  (let* ((field-list (clean-field-list field-list)) ; remove lone comments
-	 (uflds (fold-right unitize-comp-decl '() (cdr field-list))))
+  (define anon-namer
+    (let ((ix 0))
+      (lambda ()
+	(set! ix (1+ ix))
+	(string-append "_" (number->string ix)))))
+
+  (define (unitize decl seed)
+    (unitize-comp-decl decl seed #:namer anon-namer))
+
+  (let* ((fields (clean-fields (cdr field-list)))
+	 (uflds (reverse (fold unitize '() fields))))
     (let loop ((decls uflds))
       (if (null? decls) '()
 	  (let* ((name (caar decls))
@@ -755,12 +771,40 @@
     (ffi:int32 . ,int32) (ffi:uint32 . ,uint32) (ffi:int64 . ,int64)
     (ffi:uint64 . ,uint64) (ffi-void* . *)))
 
-(define (mtail->ffi-desc mdecl-tail)
-  (if (and (pair? mdecl-tail) (string? (car mdecl-tail))) (error "xxx"))
+(define (make-compound-stub size align)
+  (case align
+    ((1) `(make-list ,(/ size 1) ffi:uint8))
+    ((2) `(make-list ,(/ size 2) ffi:uint16))
+    ((4) `(make-list ,(/ size 4) ffi:uint32))
+    ((8) `(make-list ,(/ size 8) ffi:uint64))
+    (else (fherr "can't deal with this alignment: ~S" align))))
+
+;; assumes fields are unitized
+(define (bounding-mtail-for-union-fields fields)
+  (let loop ((btail #f) (mxsz 0) (mxal 0) (flds fields))
+    (if (null? flds) btail
+	(let ((mtail (cdr (udecl->mdecl (car flds)))))
+	  (call-with-values (lambda () (sizeof-mtail mtail #f)) ;; #f is arch
+	    (lambda (sz al)
+	      (if (> sz mxsz)
+		  (loop mtail sz (max al mxal) (cdr flds))
+		  (loop btail mxsz (max al mxal) (cdr flds)))))))))
+
+;;(else (fherr "can't deal with this alignment: ~S" align))))
+
+(define (clean-and-unitize-fields fields)
+  (map cdr (fold-right unitize-decl '() (clean-fields fields))))
+
+;;(define* (mtail->ffi-desc mdecl-tail #:optional in-compound)
+(export mtail->ffi-desc)
+(define* (mtail->ffi-desc mdecl-tail #:optional in-compound)
+
+  (define (eval-dim dim) ;; may want to catch errors
+    (eval-c99-cx dim (*udict*) (*ddict*)))
+
+  ;;(if (and (pair? mdecl-tail) (string? (car mdecl-tail))) (error "xxx"))
   (pmatch mdecl-tail
     (((pointer-to) . ,rest) 'ffi-void*)
-    (((array-of) . ,rest) 'ffi-void*)
-    (((array-of ,size) . ,rest) 'ffi-void*)
     (((fixed-type ,name))
      (or (assoc-ref ffi-typemap name)
 	 (fherr/once "no FFI type for ~A" name)))
@@ -774,6 +818,14 @@
     (((enum-def . ,rest2) . ,rest1) 'ffi:int)
     (((enum-ref . ,rest2) . ,rest1) 'ffi:int)
 
+    (((array-of ,dim) . ,rest)
+     (if in-compound
+	 (let ((dim (eval-dim dim)))
+	   `(make-list ,dim ,(mtail->ffi-desc rest)))
+	 'ffi-void*))
+    (((array-of) . ,rest)
+     (mtail->ffi-desc `((array-of "0") . ,rest) in-compound))
+    
     (((struct-def (field-list . ,fields)))
      `(list ,@(map (lambda (fld)
 		     (let* ((udict (unitize-comp-decl fld))
@@ -781,34 +833,22 @@
 			    (udecl (cdar udict))
 			    (udecl (udecl-rem-type-qual udecl))
 			    (mdecl (udecl->mdecl udecl)))
-		       (mtail->ffi-desc (cdr mdecl))))
-		   fields)))
+		       (mtail->ffi-desc (cdr mdecl) #t)))
+		   (clean-and-unitize-fields fields))))
     (((struct-def (ident ,name) ,field-list))
      (mtail->ffi-desc `((struct-def ,field-list))))
     
     (((union-def (field-list . ,fields)))
-     ;; TODO check libffi on how unions are passed and returned.
-     ;; I assume here never passed as a floating point.
-     ;; This should use bounding-struct-descriptor from bytestructures
-     (let loop ((type #f) (size 0) (flds fields))
-       (if (null? flds)
-	   (case type ((double) 'ffi:uint64) ((float) 'ffi:uint32) (else type))
-	   (let* ((udict (unitize-comp-decl (car flds)))
-		  (udecl (cdar udict))
-		  (udecl (udecl-rem-type-qual udecl))
-		  (mdecl (udecl->mdecl udecl))
-		  (ftype (mtail->ffi-desc (cdr mdecl)))
-		  (ftval (assq-ref ffi-symmap ftype))
-		  (fsize (sizeof ftval)))
-	     (if (> fsize size)
-		 (loop ftype fsize (cdr flds))
-		 (loop type size (cdr flds)))))))
+     (mtail->ffi-desc
+      (bounding-mtail-for-union-fields
+       (clean-and-unitize-fields fields))
+      #t))
     (((union-def (ident ,name) ,field-list))
      (mtail->ffi-desc `((union-def ,field-list))))
     
     (,otherwise
      (sferr "mtail->ffi-desc missed:\n") (pperr mdecl-tail)
-     (error "") (fherr "mtail->ffi-desc missed: ~S" mdecl-tail))))
+     (fherr "mtail->ffi-desc missed: ~S" mdecl-tail))))
 
 ;; Return a mdecl for the return type.  The variable is called @code{NAME}.
 (define (gen-decl-return udecl)
@@ -906,6 +946,18 @@
 	 (string->symbol (string-append "unwrap-enum-" name)))
 	(else 'unwrap-enum)))
 
+      (((struct-ref (ident ,name)))
+       (cond
+	((member (w/struct name) wrapped)
+	 `(fht-unwrap ,(string->symbol (sw/struct name))))
+	(else #f)))
+
+      (((union-ref (ident ,name)))
+       (cond
+	((member (w/union name) wrapped)
+	 `(fht-unwrap ,(string->symbol (sw/union name))))
+	(else #f)))
+
       (((pointer-to) (typename ,typename))
        (cond
 	((member (w/* typename) defined)
@@ -926,6 +978,15 @@
 	 `(fht-unwrap ,(strings->symbol "struct-" struct-name "*")))
 	(else 'unwrap~pointer)))
 
+      ;; use this?
+      #;(((pointer-to) (union-ref (ident ,union-name) . ,rest))
+       (cond
+	((member (w/union* union-name) defined)
+	 `(fht-unwrap ,(strings->symbol "union-" union-name "*")))
+	((member (w/union union-name) defined)
+	 `(fht-unwrap ,(strings->symbol "union-" union-name "*")))
+	(else 'unwrap~pointer)))
+
       (((pointer-to) (function-returning (param-list . ,params)) . ,rest)
        (let* ((udecl (mdecl->udecl (cons "~ret" rest)))
 	      (udecl (expand-typerefs udecl (*udict*) ffi-defined))
@@ -945,7 +1006,6 @@
 
       (,otherwise
        (sferr "mdecl->fh-unwrapper missed:\n") (pperr mdecl)
-       (error "MDECL-FH-UNWRAPPER")
        (fherr "mdecl->fh-unwrapper missed: ~S" mdecl)))))
 
 (define (mdecl->fh-wrapper mdecl)
@@ -1058,13 +1118,13 @@
 	  (let ((,~name (fh-link-proc
 			 ,decl-return ,name
 			 (append (list ,@decl-params) (map car ~rest))
-			 ,(link-libs)))
+			 (force ,(link-libs))))
 		,@(gen-exec-unwrappers exec-params))
 	    ,(if exec-return (list exec-return va-call) va-call)))))
      (#f ;; separate ~name and name defines
       (ppscm `(define ,~name
 		(delay (fh-link-proc ,decl-return ,name (list ,@decl-params)
-				     ,(link-libs)))))
+				     (force ,(link-libs))))))
       (ppscm
        `(define (,sname ,@(gen-exec-arg-names exec-params))
 	  (let ,(gen-exec-unwrappers exec-params)
@@ -1074,7 +1134,7 @@
        `(define ,sname
 	  (let ((,~name
 		 (delay (fh-link-proc ,decl-return ,name (list ,@decl-params)
-				      ,(link-libs)))))
+				      (force ,(link-libs))))))
 	    (lambda ,(gen-exec-arg-names exec-params)
 	      (let ,(gen-exec-unwrappers exec-params)
 		,(if exec-return (list exec-return call) call))))))))
@@ -1087,28 +1147,38 @@
     (sfscm ";;   (~A val) => bytestructure-set!\n" name)
     (ppscm
      `(define-public ,(string->symbol name)
-	(let ((x-var (delay (fh-link-extern ,name ,desc ,(link-libs)))))
+	(let ((x-var (delay (fh-link-extern ,name ,desc (force ,(link-libs))))))
 	  (case-lambda
 	    (() (bytestructure-ref (force x-var)))
 	    ((var) (bytestructure-set! (force x-var) var))))))))
 
+(define (node-not-typeof? crit)
+  (let ((pred (node-typeof? crit)))
+    (lambda (node) (not (pred node)))))
+
 ;; assume unit-declarator
-;; TODO (ptr-declr (pointer (type-qual-list (type-qual "const"))))
 ;; See also stripdown-specl and stripdown-declr in @file{munge.scm}.
-(define (cleanup-udecl specl declr)
-  (let* ((fctn? (pair? ((sxpath '(// ftn-declr)) declr)))
-	 (specl (remove (lambda (node)
-			  (or (equal? node '(stor-spec (auto)))
-			      (equal? node '(stor-spec (register)))
-			      (equal? node '(stor-spec (static)))
-			      (and (pair? node)
-				   (equal? (car node) 'type-qual))))
+;; and make this more efficient
+(export cleanup-udecl)
+(define cleanup-udecl
+  (let* ((ftn-sel (node-closure (node-typeof? 'ftn-declr)))
+	 (fctn? (lambda (n) (pair? (ftn-sel n))))
+	 (cruft (node-self (node-not-typeof? 'type-qual))))
+    (lambda (specl declr)
+      (let* ((specl (remove (lambda (node)
+			      (or (equal? node '(stor-spec (auto)))
+				  (equal? node '(stor-spec (register)))
+				  (equal? node '(stor-spec (static)))
+				  (equal? node '(type-qual (volatile)))
+				  (equal? node '(type-qual (restrict)))))
+			    specl))
+	     (specl (if (fctn? declr)
+			(remove (lambda (node)
+				  (equal? node '(stor-spec (extern)))) specl)
 			specl))
-	 (specl (if fctn?
-		    (remove (lambda (node)
-			      (equal? node '(stor-spec (extern)))) specl)
-		    specl)))
-    (values specl declr)))
+	     ;; remove cruft like attributes and asms and initizers)
+	     (declr (and declr (sx-list (sx-tag declr) #f (sx-ref declr 1)))))
+	(values specl declr)))))
 
 ;; @deffn {Procecure} back-ref-extend! decl typename
 ;; @deffnx {Procecure} back-ref-getall decl typename
@@ -1127,6 +1197,9 @@
 
 ;;^-- instead have user run (ref<=>deref! ...
 
+;; This needs a total overhaul:
+;; I should be processing mdecl's instead of udecl's.
+
 ;; @deffn {Procedure} cnvt-udecl udecl udict wrapped defined)
 ;; Given udecl produce a ffi-spec.
 ;; Return updated (string based) keep-list, which will be modified if the
@@ -1139,6 +1212,8 @@
 ;; for any type we also declare a poitner type
 ;; TODO: decls need to be broken out into one of the forms:
 ;; function typedef struct-ref/def union-ref/def enum variable
+;; TODO: should I just remove "extern" altogether and assume
+;; variables are extern anyhow?
 (define (cnvt-udecl udecl udict wrapped defined)
   ;; This is a bit sloppy in that we have to know if the converters are
   ;; creating wrappers and/or (type) defines.
@@ -1199,11 +1274,21 @@
 	 (stor-spec (typedef))
 	 (type-spec (fixed-type ,name)))
 	(init-declr (ident ,typename)))
-       ;; FIX
        (sfscm "(define-public ~A-desc ~A)\n"
 	      typename (assoc-ref bs-typemap name))
        (values wrapped defined))
 
+      ;; typedef int *foo_t;
+      ((udecl
+	(decl-spec-list
+	 (stor-spec (typedef))
+	 (type-spec (fixed-type ,name)))
+	(init-declr (ptr-declr (pointer) (ident ,typename))))
+       (sfscm "(define-public ~A-desc (fh:pointer ~A))\n"
+	      typename (assoc-ref bs-typemap name))
+       (fhscm-def-pointer typename)
+       (values wrapped defined))
+
       ;; typedef double foo_t;
       ;; If fh-object? then should be bytestructure.
       ;; Should wrap be to number or bytestructure?
@@ -1240,6 +1325,26 @@
        (fhscm-def-pointer typename)
        (values (cons typename wrapped) (cons typename defined)))
 
+      ;; typedef foo_t foo_array_t[123];
+      ((udecl
+	(decl-spec-list (stor-spec (typedef)) (type-spec (typename ,name)))
+	(init-declr (ary-declr (ident ,typename) (p-expr (fixed ,size)))))
+       (sferr "ffi-help/cnvt-udecl in-work: typedef type array\n")
+       ;;(pretty-print-c99 udecl)
+       ;;(pperr udecl)
+       (let* ((eltname (rename name))
+	      (name (rename typename))
+	      (st-name (if (string? name) name (symbol->string name)))
+	      (sy-name (if (string? name) (string->symbol name) name))
+	      (elt-desc (string->symbol (string-append eltname "-desc")))
+	      (pred (string->symbol (string-append st-name "?")))
+	      (make (string->symbol (string-append "make-" st-name))))
+	 (sfscm "(define-public ~A-desc (bs:vector ~A-desc ~A))\n"
+		name eltname size)
+	 (upscm `(define-fh-vector-type ,sy-name ,elt-desc ,pred ,make))
+	 (fhscm-export-def name))
+       (values (cons typename wrapped) (cons typename defined)))
+
       ;; typedef enum foo { ... } foo_t;
       ((udecl
 	(decl-spec-list
@@ -1732,10 +1837,9 @@
       ;; === missed =====================
 
       (,otherwise
-       (sferr "ffi-help/cnvt-udecl misssed:\n")
-       (pretty-print-c99 udecl)
-       (pperr udecl)
-       (quit)
+       (sferr "ffi-help/cnvt-udecl missed:\n")
+       (pretty-print-c99 clean-udecl)
+       (pperr clean-udecl)
        (values wrapped defined)))))
 
 
@@ -1745,12 +1849,10 @@
 ;; replacemnts down to constants (strings, integers, etc)
 (define (gen-lookup-proc mod-def-names udict ddict ext-mods)
 
-  (define err-port (open-output-file "/dev/null"))
-
   (define typerefs (udict->typedef-names udict))
 
   (define (eval-code-string code)
-    (let* ((tree (with-error-to-port err-port
+    (let* ((tree (with-error-to-port (dev/null)
 		   (lambda () (parse-c99x code typerefs))))
 	   (value (and tree (eval-c99-cx tree udict ddict))))
       value))
@@ -1789,9 +1891,7 @@
 	 ((number? obj) obj)
 	 ((symbol? obj) (,sv-name obj))
 	 ((fh-object? obj) (struct-ref obj 0)) ;; ???
-	 (else (error "type mismatch")))))
-    
-    (close err-port)))
+	 (else (error "type mismatch")))))))
 
 
 ;; === Parsing the C header(s)
@@ -1917,7 +2017,7 @@
 	  (fold
 	   (lambda (upath seed)
 	     (unless (resolve-module upath)
-	       (error "module not defined:" upath))
+	       (fherr "module not defined: ~S" upath))
 	     (let* ((modul (resolve-module upath #:ensure #f))
 		    (pname (m-path->name upath))
 		    (vname (string->symbol (string-append pname "-types")))
@@ -1925,6 +2025,9 @@
 	       (append var seed)))
 	   '() ext-mods)))
 
+    ;;(sferr "ffi-decls:\n") (pperr ffi-decls)
+    ;;(sferr "QUIT\n") (quit)
+
     ;; set globals
     (*prefix* (m-path->name path))
     (*udict* udict)
@@ -2173,7 +2276,7 @@
 		 (*renamer* identity) (*errmsgs* '()) (*prefix* "")
 		 (*mport* #t) (*udict* '()) (*ddict* '()))
     (if (not (access? filename R_OK))
-	(throw 'ffi-help-error "ERROR: not found: ~S" filename))
+	(fherr "ERROR: not found: ~S" filename))
     (call-with-input-file filename
       (lambda (iport)
 	(let ((output (or (assq-ref 'output options)
