@@ -1,6 +1,6 @@
 ;;; nyacc/lang/c99/pprint.scm - C pretty-printer
 
-;; Copyright (C) 2015-2018 Matthew R. Wette
+;; Copyright (C) 2015-2018,2021 Matthew R. Wette
 ;;
 ;; This library is free software; you can redistribute it and/or
 ;; modify it under the terms of the GNU Lesser General Public
@@ -345,8 +345,13 @@
 	   (else (ppx (car pair)))))
 	expr-list))
 
+      ((stmt-expr (block-item-list . ,items))
+       (sf "({\n") (push-il) (for-each ppx items) (pop-il) (sf "})"))
+
       ((udecl . ,rest)
        (ppx `(decl . ,rest)))
+      ((decl (@ . ,attr))		; GNU extension
+       (sf ";"))
       ((decl (@ . ,attr) ,decl-spec-list)
        (ppx decl-spec-list) (sf ";") (comm+nl attr))
       ((decl (@ . ,attr) ,decl-spec-list ,init-declr-list)
@@ -370,6 +375,8 @@
 	    ((type-qual) (sf "~A" (sx-ref (car dsl) 1)))
 	    ((fctn-spec) (sf "~A" (sx-ref (car dsl) 1)))
 	    ((type-spec) (ppx (car dsl)))
+	    ((typeof-expr typeof-type)
+	     (sf "typeof(") (ppx (sx-ref (car dsl) 1)) (sf ") "))
 	    (else (sf "[?:~S]" (car dsl))))
 	  (if (pair? (cdr dsl)) (sf " ")))
 	dsl))
@@ -388,11 +395,11 @@
 	rest))
 
       ((init-declr ,declr ,item2 ,item3) (ppx declr) (ppx item2) (ppx item3))
-      ((init-declr ,declr ,item2) (ppx declr) (ppx item2))
+      ((init-declr ,declr ,item2) (ppx declr) (sf " ") (ppx item2))
       ((init-declr ,declr) (ppx declr))
-      ((comp-declr ,declr ,item2) (ppx declr) (ppx item2))
+      ((comp-declr ,declr ,item2) (ppx declr) (sf " ") (ppx item2))
       ((comp-declr ,declr) (ppx declr))
-      ((param-declr ,declr ,item2) (ppx declr) (ppx item2))
+      ((param-declr ,declr ,item2) (ppx declr) (sf " ") (ppx item2))
       ((param-declr ,declr) (ppx declr))
       ((param-declr))
 
@@ -412,7 +419,7 @@
 	 ((enum-def) (ppx arg))
 	 ((typename) (sf "~A" (sx-ref arg 1)))
 	 ((void) (sf "void"))
-	 (else (error "missing " arg))))
+	 (else (throw 'nyacc-error "pprint: type-spec: ~S" arg))))
 
       ((struct-ref (ident ,name)) (sf "struct ~A" name))
       ((union-ref (ident ,name)) (sf "union ~A" name))
@@ -440,6 +447,11 @@
        (for-each ppx defns)
        (pop-il) (sf "}"))
 
+      ((enum-defn (@ . ,attr) (ident ,name) (attribute-list . ,attrs) ,expr)
+       (sf "~A " name) (ppx (sx-ref tree 2))
+       (sf " = ") (ppx expr) (sf ",") (comm+nl attr))
+      ((enum-defn (@ . ,attr) (ident ,name) (attribute-list . ,attrs))
+       (sf "~A " name) (ppx (sx-ref tree 2)) (sf ",") (comm+nl attr))
       ((enum-defn (@ . ,attr) (ident ,name) ,expr)
        (sf "~A = " name) (ppx expr) (sf ",") (comm+nl attr))
       ((enum-defn (@ . ,attr) (ident ,name))
@@ -482,6 +494,8 @@
 	(lambda (dsl)
 	  (case (sx-tag (car dsl))
 	    ((type-qual) (sf "~A" (sx-ref (car dsl) 1)))
+	    ((typeof-expr typeof-type)
+	     (sf "typeof(") (ppx (sx-ref (car dsl) 1)) (sf ") "))
 	    (else (sf "[?:~S]" (car dsl))))
 	  (if (pair? (cdr dsl)) (sf " ")))
 	tql))
@@ -535,7 +549,7 @@
 
       ;; initializer
       ((initzer ,expr)
-       (sf " = ") (ppx expr))
+       (sf "= ") (ppx expr))
       
       ;; initializer-list
       ((initzer-list . ,items)
@@ -633,6 +647,47 @@
        (ppx expr1) (sf "; ") (ppx expr2) (sf "; ") (ppx expr3)
        (sf ") ") (ppx stmt))
 
+      ;; asm - parser does not preserve specifiers
+      ((asm-expr (@ . ,attr) ,pat ,outs ,ins ,clobs ,gotos)
+       (if (assq-ref attr 'goto) (sf "asm goto (") (sf "asm ("))
+       (ppx pat) (ppx outs) (ppx ins) (ppx clobs) (ppx gotos) (sf ")"))
+      ((asm-expr (@ . ,attr) ,pat ,outs ,ins ,clobs)
+       (if (assq-ref attr 'volatile) (sf "asm volatile (") (sf "asm ("))
+       (ppx pat) (ppx outs) (ppx ins) (ppx clobs) (sf ")"))
+      ((asm-expr (@ . ,attr) ,pat ,outs ,ins)
+       (if (assq-ref attr 'volatile) (sf "asm volatile (") (sf "asm ("))
+       (ppx pat) (ppx outs) (ppx ins) (sf ")"))
+      ((asm-expr (@ . ,attr) ,pat ,outs)
+       (if (assq-ref attr 'volatile) (sf "asm volatile (") (sf "asm ("))
+       (ppx pat) (ppx outs) (sf ")"))
+      ((asm-expr (@ . ,attr) ,pat)
+       (if (assq-ref attr 'volatile) (sf "asm volatile (") (sf "asm ("))
+       (ppx pat) (sf ")"))
+      ((asm-outputs . ,elts)
+       (sf ": ")
+       (pair-for-each
+	(lambda (pair) (ppx (car pair)) (if (pair? (cdr pair)) (sf ", ")))
+	elts))
+      ((asm-inputs . ,elts)
+       (sf ": ")
+       (pair-for-each
+	(lambda (pair) (ppx (car pair)) (if (pair? (cdr pair)) (sf ", ")))
+	elts))
+      ((asm-clobbers . ,elts)
+       (sf ": ")
+       (pair-for-each
+	(lambda (pair) (ppx (car pair)) (if (pair? (cdr pair)) (sf ", ")))
+	elts))
+      ((asm-operand (ident ,name) ,str ,arg)
+       (sf "[~A] " name) (ppx str) (sf " (") (ppx arg) (sf ")"))
+      ((asm-operand ,str ,arg)
+       (ppx str) (sf " (") (ppx arg) (sf ")"))
+      ((asm-gotos . ,elts)
+       (sf ": ")
+       (pair-for-each
+	(lambda (pair) (ppx (car pair)) (if (pair? (cdr pair)) (sf ", ")))
+	elts))
+       
       ;; jump-statement
       ((goto ,where)
        (pop-il)			; unindent
@@ -649,15 +704,16 @@
       ((trans-unit . ,items)
        (pair-for-each
 	(lambda (pair)
-	  (let ((this (car pair)) (next (and (pair? (cdr pair)) (cadr pair))))
+	  (let ((this (car pair)) (this-tag (caar pair))
+		(next-tag (and (pair? (cdr pair)) (caadr pair))))
 	    (ppx this)
-	    (cond ;; add blank line if next is different or fctn defn
-	     ((not next))
-	     ((eqv? (sx-tag this) (sx-tag next)))
-	     ((eqv? (sx-tag this) 'comment))
-	     ((eqv? (sx-tag next) 'comment) (sf "\n"))
-	     ((not (eqv? (sx-tag this) (sx-tag next))) (sf "\n"))
-	     ((eqv? (sx-tag next) 'fctn-defn) (sf "\n")))))
+	    (cond ;; heuristics for adding blank lines:
+	     ((not next-tag))
+	     ((eqv? this-tag 'comment))
+	     ((eqv? next-tag 'fctn-defn) (sf "\n"))
+	     ((eqv? this-tag next-tag))
+	     ((eqv? next-tag 'comment) (sf "\n"))
+	     (else (sf "\n")))))
 	items))
 
       ((fctn-defn . ,rest) ;; but not yet (knr-fctn-defn)
@@ -707,7 +763,7 @@
        (pretty-print tree #:per-line-prefix "  ")
        )))
 
-  (if (not (pair? tree)) (error "expecing sxml tree"))
+  (if (not (pair? tree)) (throw 'c99-error "pprint: expecting sxml tree"))
   (ppx tree)
   (if ugly (newline)))
 
